import asyncio


class WebSocket:
    def __init__(self, **kwargs):
        self.__debug: bool = kwargs.get("debug", False)

        self.__port: int = kwargs.get("port", 8888)
        self.__ip: str = kwargs.get("ip", "127.0.0.1")

        self.__Reader, self.__Writer = None, None
        self.__func = kwargs.get("func", None)

    def send(self, data):
        asyncio.run(self.__send(data))

    def __debugPr(self, Input: str):
        if self.__debug is True:
            print(f"[DEBUG] :: {Input}")

    async def __send(self, data):
        self.__debugPr(f"Opening connection through asyncio on: {self.__ip, self.__port}")
        self.__Reader, self.__Writer = await asyncio.open_connection(self.__ip, self.__port)

        self.__Writer.write(f"{data}".encode())
        await self.__Writer.drain()
        self.__debugPr(f"Send message '{data}' to connection")

        if self.__func:
            _data = await self.__Reader.read(100)
            if _data:
                self.__debugPr(f"Message received: '{_data}'")
                retr = self.__func(_data)
                if retr:
                    self.__debugPr(f"Send message '{retr}' to connection")
                    self.__Writer.write(bytes(retr, "utf-8"))

        self.__Writer.close()


# Cl = WebSocket()
# Cl.send("Test")



from __future__ import annotations
import socket
from threading import Thread
import json


class WebSocket:
    def __init__(self, **kwargs) -> WebSocket:
        self.__debug: bool = kwargs.get("debug", False)

        self.__port = int(kwargs.get("port", 5555))
        self.__Host = "0.0.0.0"
        self.__buffer = int(kwargs.get("buffer", 1024))

        self.__Socket = socket.socket()
        self.__bind()

        self.__Clients = []
        self.__identification = kwargs.get("identifier", None)

    def __del__(self):
        self.__Socket.close()

    def __bind(self) -> socket:
        self.__Socket.bind((self.__Host, self.__port))

    def listen(self) -> Thread:
        self.__Socket.listen(1)

        while True:
            conn, addr = self.__Socket.accept()
            thread = Thread(target=self.__flag, args=[conn, addr])
            thread.start()

    def __remove(self, client: tuple):
        try:
            self.__Clients.remove(client)
        except Exception:
            return

    def __debugPr(self, Input: str):
        if self.__debug is True:
            print(f"[DEBUG] :: {Input}")

    def __flag(self, conn, addr):
        if self.__identification:
            try:
                msg = conn.recv(self.__buffer)
            except socket.error():
                pass
            msg = msg.decode()

            self.__debugPr(f"Message from address {addr} received: '{msg}'")

            retr = self.__identification(msg)

            if retr[0] is False:
                conn.send("!DISCONNECT".encode())
                self.__remove((conn, addr))
                conn.close()
                self.__debugPr(f"Closed connection to address: {addr}")
                return

        self.__Clients.append((conn, addr))

        while True:
            try:
                msg = conn.recv(self.__buffer)
            except socket.error():
                break
            msg = msg.decode()
            if msg == "!DISCONNECT":
                self.__remove((conn, addr))
                conn.close()
                self.__debugPr(f"Closed connection to address: {addr}")
                break

            for client in self.__Clients:
                if self.__identification:
                    client[0].send(json.dumps({'Address': addr, 'User': retr[1], 'Message': msg}).encode())
                else:
                    client[0].send(json.dumps({'Address': addr, 'Message': msg}).encode())

                self.__debugPr(f"Send message to address: {client[1]}")


# Hub = WebSocket()
# Hub.listen()



import asyncio


class WebSocket:
    def __init__(self, **kwargs):
        self.__debug: bool = kwargs.get("debug", False)

        self.__port: int = kwargs.get("port", 8888)
        self.__ip: str = kwargs.get("ip", "127.0.0.1")

        self.__Reader, self.__Writer = None, None

        self.__func = kwargs.get("func", None)
        self.__Server = None

    def __debugPr(self, Input: str):
        if self.__debug is True:
            print(f"[DEBUG] :: {Input}")

    def listen(self):
        asyncio.run(self.__listeningPrep())

    async def __listeningPrep(self):
        self.__debugPr(f"Starting server on: {self.__ip, self.__port}")
        self.__Server = await asyncio.start_server(self.__listen, self.__ip, self.__port)

        async with self.__Server:
            await self.__Server.serve_forever()

    async def __listen(self, reader, writer):
        addr = writer.get_extra_info('peername')

        data = await reader.read(100)
        data = data.decode()
        self.__debugPr(f"Message from address {addr} received: '{data}'")

        if self.__func:
            retr = self.__func(data)

            if retr:
                writer.write(f"{retr}".encode())
                await writer.drain()
                self.__debugPr(f"Send message '{retr}' to address: {addr}")



"""
pip._vendor is for vendoring dependencies of pip to prevent needing pip to
depend on something external.

Files inside of pip._vendor should be considered immutable and should only be
updated to versions from upstream.
"""
from __future__ import absolute_import

import glob
import os.path
import sys

# Downstream redistributors which have debundled our dependencies should also
# patch this value to be true. This will trigger the additional patching
# to cause things like "six" to be available as pip.
DEBUNDLED = False

# By default, look in this directory for a bunch of .whl files which we will
# add to the beginning of sys.path before attempting to import anything. This
# is done to support downstream re-distributors like Debian and Fedora who
# wish to create their own Wheels for our dependencies to aid in debundling.
WHEEL_DIR = os.path.abspath(os.path.dirname(__file__))


# Define a small helper function to alias our vendored modules to the real ones
# if the vendored ones do not exist. This idea of this was taken from
# https://github.com/kennethreitz/requests/pull/2567.
def vendored(modulename):
    vendored_name = "{0}.{1}".format(__name__, modulename)

    try:
        __import__(modulename, globals(), locals(), level=0)
    except ImportError:
        # We can just silently allow import failures to pass here. If we
        # got to this point it means that ``import pip._vendor.whatever``
        # failed and so did ``import whatever``. Since we're importing this
        # upfront in an attempt to alias imports, not erroring here will
        # just mean we get a regular import error whenever pip *actually*
        # tries to import one of these modules to use it, which actually
        # gives us a better error message than we would have otherwise
        # gotten.
        pass
    else:
        sys.modules[vendored_name] = sys.modules[modulename]
        base, head = vendored_name.rsplit(".", 1)
        setattr(sys.modules[base], head, sys.modules[modulename])


# If we're operating in a debundled setup, then we want to go ahead and trigger
# the aliasing of our vendored libraries as well as looking for wheels to add
# to our sys.path. This will cause all of this code to be a no-op typically
# however downstream redistributors can enable it in a consistent way across
# all platforms.
if DEBUNDLED:
    # Actually look inside of WHEEL_DIR to find .whl files and add them to the
    # front of our sys.path.
    sys.path[:] = glob.glob(os.path.join(WHEEL_DIR, "*.whl")) + sys.path

    # Actually alias all of our vendored dependencies.
    vendored("appdirs")
    vendored("cachecontrol")
    vendored("certifi")
    vendored("colorama")
    vendored("distlib")
    vendored("distro")
    vendored("html5lib")
    vendored("six")
    vendored("six.moves")
    vendored("six.moves.urllib")
    vendored("six.moves.urllib.parse")
    vendored("packaging")
    vendored("packaging.version")
    vendored("packaging.specifiers")
    vendored("pep517")
    vendored("pkg_resources")
    vendored("progress")
    vendored("requests")
    vendored("requests.exceptions")
    vendored("requests.packages")
    vendored("requests.packages.urllib3")
    vendored("requests.packages.urllib3._collections")
    vendored("requests.packages.urllib3.connection")
    vendored("requests.packages.urllib3.connectionpool")
    vendored("requests.packages.urllib3.contrib")
    vendored("requests.packages.urllib3.contrib.ntlmpool")
    vendored("requests.packages.urllib3.contrib.pyopenssl")
    vendored("requests.packages.urllib3.exceptions")
    vendored("requests.packages.urllib3.fields")
    vendored("requests.packages.urllib3.filepost")
    vendored("requests.packages.urllib3.packages")
    vendored("requests.packages.urllib3.packages.ordered_dict")
    vendored("requests.packages.urllib3.packages.six")
    vendored("requests.packages.urllib3.packages.ssl_match_hostname")
    vendored("requests.packages.urllib3.packages.ssl_match_hostname."
             "_implementation")
    vendored("requests.packages.urllib3.poolmanager")
    vendored("requests.packages.urllib3.request")
    vendored("requests.packages.urllib3.response")
    vendored("requests.packages.urllib3.util")
    vendored("requests.packages.urllib3.util.connection")
    vendored("requests.packages.urllib3.util.request")
    vendored("requests.packages.urllib3.util.response")
    vendored("requests.packages.urllib3.util.retry")
    vendored("requests.packages.urllib3.util.ssl_")
    vendored("requests.packages.urllib3.util.timeout")
    vendored("requests.packages.urllib3.util.url")
    vendored("resolvelib")
    vendored("tenacity")
    vendored("toml")
    vendored("toml.encoder")
    vendored("toml.decoder")
    vendored("urllib3")



import logging
import pathlib
import sys
import sysconfig
from typing import List, Optional

from pip._internal.models.scheme import SCHEME_KEYS, Scheme

from . import _distutils, _sysconfig
from .base import (
    USER_CACHE_DIR,
    get_major_minor_version,
    get_src_prefix,
    site_packages,
    user_site,
)

__all__ = [
    "USER_CACHE_DIR",
    "get_bin_prefix",
    "get_bin_user",
    "get_major_minor_version",
    "get_platlib",
    "get_prefixed_libs",
    "get_purelib",
    "get_scheme",
    "get_src_prefix",
    "site_packages",
    "user_site",
]


logger = logging.getLogger(__name__)


def _default_base(*, user: bool) -> str:
    if user:
        base = sysconfig.get_config_var("userbase")
    else:
        base = sysconfig.get_config_var("base")
    assert base is not None
    return base


def _warn_if_mismatch(old: pathlib.Path, new: pathlib.Path, *, key: str) -> bool:
    if old == new:
        return False
    issue_url = "https://github.com/pypa/pip/issues/9617"
    message = (
        "Value for %s does not match. Please report this to <%s>"
        "\ndistutils: %s"
        "\nsysconfig: %s"
    )
    logger.debug(message, key, issue_url, old, new)
    return True


def _log_context(
    *,
    user: bool = False,
    home: Optional[str] = None,
    root: Optional[str] = None,
    prefix: Optional[str] = None,
) -> None:
    message = (
        "Additional context:" "\nuser = %r" "\nhome = %r" "\nroot = %r" "\nprefix = %r"
    )
    logger.debug(message, user, home, root, prefix)


def get_scheme(
    dist_name,  # type: str
    user=False,  # type: bool
    home=None,  # type: Optional[str]
    root=None,  # type: Optional[str]
    isolated=False,  # type: bool
    prefix=None,  # type: Optional[str]
):
    # type: (...) -> Scheme
    old = _distutils.get_scheme(
        dist_name,
        user=user,
        home=home,
        root=root,
        isolated=isolated,
        prefix=prefix,
    )
    new = _sysconfig.get_scheme(
        dist_name,
        user=user,
        home=home,
        root=root,
        isolated=isolated,
        prefix=prefix,
    )

    base = prefix or home or _default_base(user=user)
    warned = []
    for k in SCHEME_KEYS:
        # Extra join because distutils can return relative paths.
        old_v = pathlib.Path(base, getattr(old, k))
        new_v = pathlib.Path(getattr(new, k))

        # distutils incorrectly put PyPy packages under ``site-packages/python``
        # in the ``posix_home`` scheme, but PyPy devs said they expect the
        # directory name to be ``pypy`` instead. So we treat this as a bug fix
        # and not warn about it. See bpo-43307 and python/cpython#24628.
        skip_pypy_special_case = (
            sys.implementation.name == "pypy"
            and home is not None
            and k in ("platlib", "purelib")
            and old_v.parent == new_v.parent
            and old_v.name == "python"
            and new_v.name == "pypy"
        )
        if skip_pypy_special_case:
            continue

        warned.append(_warn_if_mismatch(old_v, new_v, key=f"scheme.{k}"))

    if any(warned):
        _log_context(user=user, home=home, root=root, prefix=prefix)

    return old


def get_bin_prefix():
    # type: () -> str
    old = _distutils.get_bin_prefix()
    new = _sysconfig.get_bin_prefix()
    if _warn_if_mismatch(pathlib.Path(old), pathlib.Path(new), key="bin_prefix"):
        _log_context()
    return old


def get_bin_user():
    # type: () -> str
    return _sysconfig.get_scheme("", user=True).scripts


def get_purelib():
    # type: () -> str
    """Return the default pure-Python lib location."""
    old = _distutils.get_purelib()
    new = _sysconfig.get_purelib()
    if _warn_if_mismatch(pathlib.Path(old), pathlib.Path(new), key="purelib"):
        _log_context()
    return old


def get_platlib():
    # type: () -> str
    """Return the default platform-shared lib location."""
    old = _distutils.get_platlib()
    new = _sysconfig.get_platlib()
    if _warn_if_mismatch(pathlib.Path(old), pathlib.Path(new), key="platlib"):
        _log_context()
    return old


def get_prefixed_libs(prefix):
    # type: (str) -> List[str]
    """Return the lib locations under ``prefix``."""
    old_pure, old_plat = _distutils.get_prefixed_libs(prefix)
    new_pure, new_plat = _sysconfig.get_prefixed_libs(prefix)

    warned = [
        _warn_if_mismatch(
            pathlib.Path(old_pure),
            pathlib.Path(new_pure),
            key="prefixed-purelib",
        ),
        _warn_if_mismatch(
            pathlib.Path(old_plat),
            pathlib.Path(new_plat),
            key="prefixed-platlib",
        ),
    ]
    if any(warned):
        _log_context(prefix=prefix)

    if old_pure == old_plat:
        return [old_pure]
    return [old_pure, old_plat]



from __future__ import annotations
import socket
from threading import Thread


class WebSocket:
    def __init__(self, **kwargs) -> WebSocket:
        self.__debug: bool = kwargs.get("debug", False)

        self.__port = int(kwargs.get("port", 5555))
        self.__Host = "0.0.0.0"
        self.__buffer = int(kwargs.get("buffer", 1024))

        self.__Socket = socket.socket()
        self.__bind()

        self.__Function = kwargs.get("func", None)

    def __debugPr(self, Input: str):
        if self.__debug is True:
            print(f"[DEBUG] :: {Input}")

    def __del__(self):
        self.__Socket.close()

    def __bind(self) -> socket:
        self.__Socket.bind((self.__Host, self.__port))

    def listen(self) -> Thread:
        self.__Socket.listen(1)

        while True:
            conn, addr = self.__Socket.accept()
            self.__debugPr(f"Starting Thread for incoming connection")
            thread = Thread(target=self.__flag, args=[conn, addr])
            thread.start()

    def __flag(self, conn, addr):
        while True:
            try:
                msg = conn.recv(self.__buffer)
            except socket.error():
                break
            msg = msg.decode()
            self.__debugPr(f"Message from address {addr} received: '{msg}'")

            if msg == "!DISCONNECT":
                self.__debugPr(f"Closed connection to address: {addr}")
                conn.close()
                break

            if self.__Function:
                retr = self.__Function(msg)
                if retr:
                    if retr != "!DISCONNECT":
                        conn.send(f"{retr}".encode())
                        self.__debugPr(f"Closed connection to address: {addr}")
                        conn.close()

                    else:

                        self.__debugPr(f"Send Message: '{retr}' to address: {addr}")
                        conn.send(f"{retr}".encode())


# Server = WebSocket()
# Server.listen()



from __future__ import annotations
import socket


class WebSocket:
    def __init__(self, **kwargs) -> WebSocket:
        self.__debug: bool = kwargs.get("debug", False)

        self.__port = int(kwargs.get("port", 5555))
        self.__Host = socket.gethostbyname(socket.gethostname())
        self.__buffer = int(kwargs.get("buffer", 1024))

        self.__Function = kwargs.get("func", None)

        self.__Socket = socket.socket()
        self.__debugPr(f"Opening connection on: {(self.__Host, self.__port)}")
        self.__Socket.connect((self.__Host, self.__port))

    def __debugPr(self, Input: str):
        if self.__debug is True:
            print(f"[DEBUG] :: {Input}")

    def __del__(self):
        self.__Socket.send("!DISCONNECT".encode())
        self.__Socket.close()

    def send(self, data) -> send:
        self.__Socket.send(f"{data}".encode())
        self.__debugPr(f"Send message '{data}' to connection")

        if self.__Function:

            rec = self.__Socket.recv(self.__buffer)
            rec = rec.decode()

            self.__Function(rec)

    def reconnect(self) -> WebSocket:
        self.__Socket.send("!DISCONNECT".encode())
        self.__Socket.close()
        self.__debugPr(f"Reconnecting Socket")
        self.__Socket = socket.socket()
        self.__Socket.connect((self.__Host, self.__port))


# Client = WebSocket()
# Client.send("Test")



from __future__ import annotations
from socket import socket

# Importing the different types of Hub_Server
from BetterSockets.Asyncio.Client import WebSocket as AsyncClient
from BetterSockets.Threads.Client import WebSocket as ThreadClient


class Client:
    def __init__(self, **kwargs) -> socket:
        self.__type: bool = kwargs.get("is_async", False)

        if self.__type is True:
            self.__Socket = AsyncClient(**kwargs)

        elif self.__type is False:
            self.__Socket = ThreadClient(**kwargs)

    def send(self, data: str):
        self.__Socket.send(data)



from __future__ import annotations
from socket import socket

# Importing the different types of Process_Server
from BetterSockets.Asyncio.Process_Server import WebSocket as AsyncSocket
from BetterSockets.Threads.Process_Server import WebSocket as ThreadSocket


class ProcessorServer:
    def __init__(self, **kwargs) -> socket:
        self.__type: bool = kwargs.get("is_async", False)

        if self.__type is True:
            self.__Socket = AsyncSocket(**kwargs)

        elif self.__type is False:
            self.__Socket = ThreadSocket(**kwargs)

    def listen(self):
        self.__Socket.listen()



from setuptools import setup, find_packages
from BetterSockets import __version__

classifiers = \
    [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers ",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only"
]

setup(
    name="BetterSockets",
    version=__version__,
    description="Better Python sockets and asyncio streams",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Drageast/BetterSockets",
    author="Luca Michael Schmidt",
    author_email="schmidt.lucamichael@gmail.com",
    license="MIT",
    classifiers=classifiers,
    packages=find_packages()
)



import logging
import re
from typing import Container, Iterator, List, Optional, Union

from pip._vendor.packaging.version import LegacyVersion, Version

from pip._internal.utils.misc import stdlib_pkgs  # TODO: Move definition here.

DistributionVersion = Union[LegacyVersion, Version]

logger = logging.getLogger(__name__)


class BaseDistribution:
    @property
    def location(self):
        # type: () -> Optional[str]
        """Where the distribution is loaded from.

        A string value is not necessarily a filesystem path, since distributions
        can be loaded from other sources, e.g. arbitrary zip archives. ``None``
        means the distribution is created in-memory.
        """
        raise NotImplementedError()

    @property
    def metadata_version(self):
        # type: () -> Optional[str]
        """Value of "Metadata-Version:" in the distribution, if available."""
        raise NotImplementedError()

    @property
    def canonical_name(self):
        # type: () -> str
        raise NotImplementedError()

    @property
    def version(self):
        # type: () -> DistributionVersion
        raise NotImplementedError()

    @property
    def installer(self):
        # type: () -> str
        raise NotImplementedError()

    @property
    def editable(self):
        # type: () -> bool
        raise NotImplementedError()

    @property
    def local(self):
        # type: () -> bool
        raise NotImplementedError()

    @property
    def in_usersite(self):
        # type: () -> bool
        raise NotImplementedError()


class BaseEnvironment:
    """An environment containing distributions to introspect."""

    @classmethod
    def default(cls):
        # type: () -> BaseEnvironment
        raise NotImplementedError()

    @classmethod
    def from_paths(cls, paths):
        # type: (Optional[List[str]]) -> BaseEnvironment
        raise NotImplementedError()

    def get_distribution(self, name):
        # type: (str) -> Optional[BaseDistribution]
        """Given a requirement name, return the installed distributions."""
        raise NotImplementedError()

    def _iter_distributions(self):
        # type: () -> Iterator[BaseDistribution]
        """Iterate through installed distributions.

        This function should be implemented by subclass, but never called
        directly. Use the public ``iter_distribution()`` instead, which
        implements additional logic to make sure the distributions are valid.
        """
        raise NotImplementedError()

    def iter_distributions(self):
        # type: () -> Iterator[BaseDistribution]
        """Iterate through installed distributions."""
        for dist in self._iter_distributions():
            # Make sure the distribution actually comes from a valid Python
            # packaging distribution. Pip's AdjacentTempDirectory leaves folders
            # e.g. ``~atplotlib.dist-info`` if cleanup was interrupted. The
            # valid project name pattern is taken from PEP 508.
            project_name_valid = re.match(
                r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$",
                dist.canonical_name,
                flags=re.IGNORECASE,
            )
            if not project_name_valid:
                logger.warning(
                    "Ignoring invalid distribution %s (%s)",
                    dist.canonical_name,
                    dist.location,
                )
                continue
            yield dist

    def iter_installed_distributions(
        self,
        local_only=True,  # type: bool
        skip=stdlib_pkgs,  # type: Container[str]
        include_editables=True,  # type: bool
        editables_only=False,  # type: bool
        user_only=False,  # type: bool
    ):
        # type: (...) -> Iterator[BaseDistribution]
        """Return a list of installed distributions.

        :param local_only: If True (default), only return installations
        local to the current virtualenv, if in a virtualenv.
        :param skip: An iterable of canonicalized project names to ignore;
            defaults to ``stdlib_pkgs``.
        :param include_editables: If False, don't report editables.
        :param editables_only: If True, only report editables.
        :param user_only: If True, only report installations in the user
        site directory.
        """
        it = self.iter_distributions()
        if local_only:
            it = (d for d in it if d.local)
        if not include_editables:
            it = (d for d in it if not d.editable)
        if editables_only:
            it = (d for d in it if d.editable)
        if user_only:
            it = (d for d in it if d.in_usersite)
        return (d for d in it if d.canonical_name not in skip)
